"""LaviRot

    .. [1] 'Desenvolvimento de uma interface gráﬁca para simulação
       de máquinas rotativas, com foco na inﬂuência de selos labirintos
       na dinâmica estrutural / Paulo Arthur Lima de Deus - Rio de
       Janeiro: UFRJ/ Escola Politécnica, 2015.
"""
from .elements import *
from .rotor import *
from .results import *

# TODO add setup.py

__version__ = '0.1.5b'
